#!/usr/bin/python
#
# Decorators.
#   - Cameron Simpson <cs@cskk.id.au> 02jul2017
#

r'''
Assorted decorator functions.
'''

from collections import defaultdict
from contextlib import contextmanager
from inspect import isgeneratorfunction
import sys
import time
import traceback
from cs.gimmicks import warning

__version__ = '20220311'

DISTINFO = {
    'keywords': ["python2", "python3"],
    'classifiers': [
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
    ],
    'install_requires': ['cs.gimmicks'],
}

def fmtdoc(func):
  ''' Decorator to replace a function's docstring with that string
      formatted against the function's module `__dict__`.

      This supports simple formatted docstrings:

          ENVVAR_NAME = 'FUNC_DEFAULT'

          @fmtdoc
          def func():
              """Do something with os.environ[{ENVVAR_NAME}]."""
              print(os.environ[ENVVAR_NAME])

      This gives `func` this docstring:

          Do something with os.environ[FUNC_DEFAULT].

      *Warning*: this decorator is intended for wiring "constants"
      into docstrings, not for dynamic values. Use for other types
      of values should be considered with trepidation.
  '''
  func.__doc__ = func.__doc__.format(**sys.modules[func.__module__].__dict__)
  return func

def decorator(deco):
  ''' Wrapper for decorator functions to support optional arguments.

      The actual decorator function ends up being called as:

          mydeco(func, *da, **dkw)

      allowing `da` and `dkw` to affect the behaviour of the decorator `mydeco`.

      Examples:

          # define your decorator as if always called with func and args
          @decorator
          def mydeco(func, *da, arg2=None):
            ... decorate func subject to the values of da and arg2

          # mydeco called with defaults
          @mydeco
          def func1(...):
            ...

          @ mydeco called with nondefault arguments
          @mydeco('foo', arg2='bah')
          def func2(...):
            ...
  '''

  def metadeco(*da, **dkw):
    ''' Compute either the wrapper function for `func`
        or a decorator expecting to get `func` when used.

        If there is at least one positional parameter
        and it is callable the it is presumed to be the function to decorate;
        decorate it directly.

        Otherwise return a decorator using the provided arguments,
        ready for the subsequent function.
    '''
    if len(da) > 0 and callable(da[0]):
      # `func` is already supplied, pop it off and decorate it now.
      func = da[0]
      da = tuple(da[1:])
      # decorate func
      decorated = deco(func, *da, **dkw)
      if decorated is not func:
        # we got a wrapper function back, pretty up the returned wrapper
        try:
          decorated.__name__ = getattr(func, '__name__', str(func))
        except AttributeError:
          pass
        if not getattr(decorated, '__doc__', None):
          decorated.__doc__ = getattr(func, '__doc__', '')
        func_module = getattr(func, '__module__', None)
        try:
          decorated.__module__ = func_module
        except AttributeError:
          pass
      return decorated

    # `func` is not supplied, collect the arguments supplied and return a
    # decorator which takes the subsequent callable and returns
    # `deco(func, *da, **kw)`.
    def overdeco(func):
      decorated = deco(func, *da, **dkw)
      decorated.__doc__ = getattr(func, '__doc__', '')
      func_module = getattr(func, '__module__', None)
      try:
        decorated.__module__ = func_module
      except AttributeError:
        pass
      return decorated

    return overdeco

  metadeco.__doc__ = getattr(deco, '__doc__', '')
  metadeco.__module__ = getattr(deco, '__module__', None)
  return metadeco

@decorator
def contextdecorator(cmgrfunc):
  ''' A decorator for a context manager function `cmgrfunc`
      which turns it into a decorator for other functions.

      This supports easy implementation of "setup" and "teardown"
      code around other functions without the tedium of defining
      the wrapper function itself. See the examples below.

      The resulting context manager accepts an optional keyword
      parameter `provide_context`, default `False`. If true, the
      context returned from the context manager is provided as the
      first argument to the call to the wrapped function.

      Note that the context manager function `cmgrfunc`
      has _not_ yet been wrapped with `@contextmanager`,
      that is done by `@contextdecorator`.

      This decorator supports both normal functions and generator functions.

      With a normal function the process is:
      * call the context manager with `(func,a,kw,*da,**dkw)`,
        returning `ctxt`,
        where `da` and `dkw` are the positional and keyword parameters
        supplied when the decorator was defined.
      * within the context
        return the value of `func(ctxt,*a,**kw)` if `provide_context` is true
        or the value of `func(*a,**kw)` if not (the default)

      With a generator function the process is:
      * obtain an iterator by calling `func(*a,**kw)`
      * for iterate over the iterator, yielding its results,
        by calling the context manager with `(func,a,kw,**da,**dkw)`,
        around each `next()`
      Note that it is an error to provide a true value for `provide_context`
      if the decorated function is a generator function.

      Some examples follow.

      Trace the call and return of a specific function:

          @contextdecorator
          def tracecall(func, a, kw):
              """ Trace the call and return from some function.
                  This can easily be adapted to purposes such as timing a
                  function call or logging use.
              """
              print("call %s(*%r,**%r)" % (func, a, kw))
              try:
                yield
              except Exception as e:
                print("exception from %s(*%r,**%r): %s" % (func, a, kw, e))
                raise
              else:
                print("return from %s(*%r,**%r)" % (func, a, kw))

          @tracecall
          def f():
              """ Some function to trace.
              """

          @tracecall(provide_context=True):
          def f(ctxt, *a, **kw):
              """ A function expecting the context object as its first argument,
                  ahead of whatever other arguments it would normally require.
              """

      See who is making use of a generator's values,
      when a generator might be invoked in one place and consumed elsewhere:

          from cs.py.stack import caller

          @contextdecorator
          def genuser(genfunc, *a, **kw):
              user = caller(-4)
              print(f"iterate over {genfunc}(*{a!r},**{kw!r}) from {user}")
              yield

          @genuser
          def linesof(filename):
              with open(filename) as f:
                  yield from f

          # obtain a generator of lines here
          lines = linesof(__file__)

          # perhaps much later, or in another function
          for lineno, line in enumerate(lines, 1):
              print("line %d: %d words" % (lineno, len(line.split())))

      Turn on "verbose mode" around a particular function:

          import sys
          import threading
          from cs.context import stackattrs

          class State(threading.local):
              def __init__(self):
                  # verbose if stderr is on a terminal
                  self.verbose = sys.stderr.isatty()

          # per thread global state
          state = State()

          @contextdecorator
          def verbose(func):
              with stackattrs(state, verbose=True) as old_attrs:
                  if not old_attrs['verbose']:
                      print(f"enabled verbose={state.verbose} for function {func}")
                  # yield the previous verbosity as the context
                  yield old_attrs['verbose']

          # turn on verbose mode
          @verbose
          def func(x, y):
              if state.verbose:
                  # print if verbose
                  print("x =", x, "y =", y)

          # turn on verbose mode and also pass in the previous state
          # as the first argument
          @verbose(provide_context=True):
          def func2(old_verbose, x, y):
              if state.verbose:
                  # print if verbose
                  print("old_verbosity =", old_verbose, "x =", x, "y =", y)
  '''
  # turn the function into a context manager
  cmgr = contextmanager(cmgrfunc)

  # prepare a new decorator which wraps functions in a context
  # manager using `cmgrfunc`
  @decorator
  def cmgrdeco(func, *da, provide_context=False, **dkw):
    ''' Decorator for functions which wraps calls to the function
        in a context manager, optionally supplying the context
        as the first argument to the called function.
    '''
    if isgeneratorfunction(func):
      if provide_context:
        raise ValueError(
            "provide_context may not be true when func:%s is a generator" %
            (func,)
        )

      def wrapped(*a, **kw):
        ''' Wrapper function:
            * obtain an iterator by calling `func(*a,**kw)`
            * iterate over the iterator, yielding its results,
              by calling the context manager with `(func,a,kw,**da,**dkw)`,
              around each `next()`
        '''
        it = func(*a, **kw)
        while True:
          with cmgr(func, a, kw, *da, **dkw):
            try:
              value = next(it)
            except StopIteration:
              break
          yield value

    else:

      def wrapped(*a, **kw):
        ''' Wrapper function:
            * call the context manager with `(func,a,kw,**da,**dkw)`,
              returning `ctxt`
            * within the context
              return the value of `func(ctxt,*a,**kw)`
              if `provide_context` is true
              or the value of `func(*a,**kw)` if not (the default)
        '''
        with cmgr(func, a, kw, *da, **dkw) as ctxt:
          if provide_context:
            a = [ctxt] + list(a)
          return func(*a, **kw)

    return wrapped

  return cmgrdeco

@decorator
def logging_wrapper(log_call, stacklevel_increment=1):
  ''' Decorator for logging call shims
      which bumps the `stacklevel` keyword argument so that the logging system
      chooses the correct frame to cite in messages.

      Note: has no effect on Python < 3.8 because `stacklevel` only
      appeared in that version.
  '''
  if (sys.version_info.major, sys.version_info.minor) < (3, 8):
    # do not wrap older Python log calls, no stacklevel keyword argument
    return log_call

  def log_func_wrapper(*a, **kw):
    stacklevel = kw.pop('stacklevel', 1)
    return log_call(*a, stacklevel=stacklevel + stacklevel_increment + 1, **kw)

  log_func_wrapper.__name__ = log_call.__name__
  log_func_wrapper.__doc__ = log_call.__doc__
  return log_func_wrapper

@decorator
def cachedmethod(
    method, attr_name=None, poll_delay=None, sig_func=None, unset_value=None
):
  ''' Decorator to cache the result of an instance or class method
      and keep a revision counter for changes.

      The cached values are stored on the instance (`self`).
      The revision counter supports the `@revised` decorator.

      This decorator may be used in 2 modes.
      Directly:

          @cachedmethod
          def method(self, ...)

      or indirectly:

          @cachedmethod(poll_delay=0.25)
          def method(self, ...)

      Optional keyword arguments:
      * `attr_name`: the basis name for the supporting attributes.
        Default: the name of the method.
      * `poll_delay`: minimum time between polls; after the first
        access, subsequent accesses before the `poll_delay` has elapsed
        will return the cached value.
        Default: `None`, meaning the value never becomes stale.
      * `sig_func`: a signature function, which should be significantly
        cheaper than the method. If the signature is unchanged, the
        cached value will be returned. The signature function
        expects the instance (`self`) as its first parameter.
        Default: `None`, meaning no signature function;
        the first computed value will be kept and never updated.
      * `unset_value`: the value to return before the method has been
        called successfully.
        Default: `None`.

      If the method raises an exception, this will be logged and
      the method will return the previously cached value,
      unless there is not yet a cached value
      in which case the exception will be reraised.

      If the signature function raises an exception
      then a log message is issued and the signature is considered unchanged.

      An example use of this decorator might be to keep a "live"
      configuration data structure, parsed from a configuration
      file which might be modified after the program starts. One
      might provide a signature function which called `os.stat()` on
      the file to check for changes before invoking a full read and
      parse of the file.

      *Note*: use of this decorator requires the `cs.pfx` module.
  '''
  from cs.pfx import Pfx  # pylint: disable=import-outside-toplevel
  if poll_delay is not None and poll_delay <= 0:
    raise ValueError("poll_delay <= 0: %r" % (poll_delay,))
  if poll_delay is not None and poll_delay <= 0:
    raise ValueError(
        "invalid poll_delay, should be >0, got: %r" % (poll_delay,)
    )

  attr = attr_name if attr_name else method.__name__
  val_attr = '_' + attr
  sig_attr = val_attr + '__signature'
  rev_attr = val_attr + '__revision'
  lastpoll_attr = val_attr + '__lastpoll'

  # pylint: disable=too-many-branches
  def cachedmethod_wrapper(self, *a, **kw):
    with Pfx("%s.%s", self, attr):
      now = None
      value0 = getattr(self, val_attr, unset_value)
      sig0 = getattr(self, sig_attr, None)
      sig = getattr(self, sig_attr, None)
      if value0 is unset_value:
        # value unknown, needs compute
        pass
      # we have a cached value for return in the following logic
      elif poll_delay is None:
        return value0
      # see if the value is stale
      lastpoll = getattr(self, lastpoll_attr, None)
      now = time.time()
      if lastpoll is not None and now - lastpoll < poll_delay:
        # reuse cache
        return value0
      # update the poll time if we use it
      setattr(self, lastpoll_attr, now)
      # check the signature if provided
      # see if the signature is unchanged
      if sig_func is not None:
        try:
          sig = sig_func(self)
        except Exception as e:  # pylint: disable=broad-except
          # signature function fails, use the cache
          warning("sig func %s(self): %s", sig_func, e, exc_info=True)
          return value0
        if sig0 is not None and sig0 == sig:
          # signature unchanged
          return value0
        # update signature
        setattr(self, sig_attr, sig)
      # compute the current value
      try:
        value = method(self, *a, **kw)
      except Exception as e:  # pylint: disable=broad-except
        # computation fails, return cached value
        if value0 is unset_value:
          # no cached value
          raise
        warning("exception calling %s(self): %s", method, e, exc_info=True)
        return value0
      # update the cache
      setattr(self, val_attr, value)
      # bump revision if the value changes
      # noncomparable values are always presumed changed
      changed = value0 is unset_value or value0 is not value
      if not changed:
        try:
          changed = value0 != value
        except TypeError:
          changed = True
      if changed:
        setattr(self, rev_attr, (getattr(self, rev_attr, None) or 0) + 1)
      return value

  ##  Doesn't work, has no access to self. :-(
  ##  # provide a .flush() function to clear the cached value
  ##  cachedmethod_wrapper.flush = lambda: setattr(self, val_attr, unset_value)

  return cachedmethod_wrapper

@decorator
def OBSOLETE(func, suggestion=None):
  ''' Decorator for obsolete functions.

      Use:

          @OBSOLETE
          def func(...):

      This emits a warning log message before calling the decorated function.
  '''

  def wrapped(*args, **kwargs):
    ''' Wrap `func` to emit an "OBSOLETE" warning before calling `func`.
    '''
    frame = traceback.extract_stack(None, 2)[0]
    caller = frame[0], frame[1]
    # pylint: disable=protected-access
    try:
      callers = func._OBSOLETE_callers
    except AttributeError:
      callers = func._OBSOLETE_callers = set()
    if caller not in callers:
      callers.add(caller)
      warning(
          "OBSOLETE call to %s:%d %s(), called from %s:%d %s",
          func.__code__.co_filename, func.__code__.co_firstlineno,
          func.__name__, frame[0], frame[1], frame[2]
      )
    return func(*args, **kwargs)

  funcname = getattr(func, '__name__', str(func))
  funcdoc = getattr(func, '__doc__', None) or ''
  doc = "OBSOLETE FUNCTION " + funcname
  if suggestion:
    doc += ' suggestion: ' + suggestion
  wrapped.__name__ = '@OBSOLETE(%s)' % (funcname,)
  wrapped.__doc__ = doc + '\n\n' + funcdoc
  return wrapped

@OBSOLETE(suggestion='cachedmethod')
def cached(*a, **kw):
  ''' Former name for @cachedmethod.
  '''
  return cachedmethod(*a, **kw)

def contextual(func):
  ''' Wrap a simple function as a context manager.

      This was written to support users of `@strable`,
      which requires its `open_func` to return a context manager;
      this turns an arbitrary function into a context manager.

      Example promoting a trivial function:

          >>> f = lambda: 3
          >>> cf = contextual(f)
          >>> with cf() as x: print(x)
          3
  '''

  @contextmanager
  def cmgr(*a, **kw):
    ''' Wrapper for `func` as a context manager.
    '''
    yield func(*a, **kw)

  func_name = getattr(func, '__name__', str(func))
  cmgr.__name__ = '@contextual(%s)' % func_name
  cmgr.__doc__ = func.__doc__
  return cmgr

@decorator
def strable(func, open_func=None):
  ''' Decorator for functions which may accept a `str`
      instead of their core type.

      Parameters:
      * `func`: the function to decorate
      * `open_func`: the "open" factory to produce the core type
        if a string is provided;
        the default is the builtin "open" function.
        The returned value should be a context manager.
        Simpler functions can be decorated with `@contextual`
        to turn them into context managers if need be.

      The usual (and default) example is a function to process an
      open file, designed to be handed a file object but which may
      be called with a filename. If the first argument is a `str`
      then that file is opened and the function called with the
      open file.

      Examples:

          @strable
          def count_lines(f):
            return len(line for line in f)

          class Recording:
            "Class representing a video recording."
            ...
          @strable(open_func=Recording)
          def process_video(r):
            ... do stuff with `r` as a Recording instance ...

      *Note*: use of this decorator requires the `cs.pfx` module.
  '''
  from cs.pfx import Pfx  # pylint: disable=import-outside-toplevel
  if open_func is None:
    open_func = open

  if isgeneratorfunction(func):

    def accepts_str(arg, *a, **kw):
      if isinstance(arg, str):
        with Pfx(arg):
          with open_func(arg) as opened:
            for item in func(opened, *a, **kw):
              yield item
      else:
        for item in func(arg, *a, **kw):
          yield item
  else:

    def accepts_str(arg, *a, **kw):
      if isinstance(arg, str):
        with Pfx(arg):
          with open_func(arg) as opened:
            return func(opened, *a, **kw)
      return func(arg, *a, **kw)

  return accepts_str

def observable_class(property_names, only_unequal=False):
  ''' Class decorator to make various instance attributes observable.

      Parameters:
      * `property_names`:
        an interable of instance property names to set up as
        observable properties. As a special case a single `str` can
        be supplied if only one attribute is to be observed.
      * `only_unequal`:
        only call the observers if the new property value is not
        equal to the previous proerty value. This requires property
        values to be comparable for inequality.
        Default: `False`, meaning that all updates will be reported.
  '''

  if isinstance(property_names, str):
    property_names = (property_names,)

  # pylint: disable=protected-access
  def make_observable_class(cls):
    ''' Annotate the class `cls` with observable properties.
    '''

    # push the per instance initialisation
    old_init = cls.__init__

    def new_init(self, *a, **kw):
      ''' New init, much like the old init...
      '''
      self._observable_class__observers = defaultdict(set)
      old_init(self, *a, **kw)

    cls.__init__ = new_init

    def add_observer(self, attr, observer):
      ''' Add an observer on `.attr` to this instance.
      '''
      self._observable_class__observers[attr].add(observer)

    cls.add_observer = add_observer

    def remove_observer(self, attr, observer):
      ''' Remove an observer on `.attr` from this instance.
      '''
      self._observable_class__observers[attr].remove(observer)

    cls.remove_observer = remove_observer

    def report_observation(self, attr):
      ''' Notify all the observers of the current value of `attr`.
      '''
      val_attr = '_' + attr
      value = getattr(self, val_attr, None)
      for observer in self._observable_class__observers[attr]:
        try:
          observer(self, attr, value)
        except Exception as e:  # pylint: disable=broad-except
          warning(
              "%s.%s=%r: observer %s(...) raises: %s",
              self,
              val_attr,
              value,
              observer,
              e,
              exc_info=True
          )

    cls.report_observation = report_observation

    def make_property(cls, attr):
      ''' Make `cls.attr` into a property which reports setattr events.
      '''
      val_attr = '_' + attr

      def getter(self):
        return getattr(self, val_attr)

      getter.__name__ = attr
      get_prop = property(getter)
      setattr(cls, attr, get_prop)

      def setter(self, new_value):
        ''' Set the attribute value and tell all the observers.
        '''
        old_value = getattr(self, val_attr, None)
        setattr(self, val_attr, new_value)
        if not only_unequal or old_value != new_value:
          self.report_observation(attr)

      setter.__name__ = attr
      set_prop = get_prop.setter(setter)
      setattr(cls, attr, set_prop)

    for property_name in property_names:
      if hasattr(cls, property_name):
        raise ValueError("%s.%s already exists" % (cls, property_name))
      make_property(cls, property_name)

    return cls

  return make_observable_class

def _teststuff():

  @contextdecorator
  def tracecall(func, a, kw):
    print("call %s(*%r,**%r)" % (func, a, kw))
    yield 9
    print("return from %s(*%r,**%r)" % (func, a, kw))

  @tracecall
  def f(*a, **kw):
    print("hello from f: a=%r, kw=%r" % (a, kw))
    return "V"

  @tracecall
  def g(r):
    yield from range(r)

  @tracecall(provide_context=True)
  def f2(ctxt, *a, **kw):
    print("hello from f2: ctxt=%s, a=%r, kw=%r" % (ctxt, a, kw))
    return "V2"

  v = f("abc", y=1)
  print("v =", v)
  v = f2("abc2", y=1)
  print("v2 =", v)
  gg = g(9)
  for i in gg:
    print("i =", i)
  sys.exit(1)

  # pylint: disable=too-few-public-methods
  class Foo:
    ''' Dummy class.
    '''

    @cachedmethod(poll_delay=2)
    def x(self, arg):
      ''' Dummy `x` method.
      '''
      return str(self) + str(arg)

  F = Foo()
  y = F.x(1)
  print("F.x() ==>", y)
  y = F.x(1)
  print("F.x() ==>", y)
  y = F.x(2)
  print("F.x() ==>", y)
  time.sleep(3)
  y = F.x(3)
  print("F.x() ==>", y)

if __name__ == '__main__':
  _teststuff()



#!/usr/bin/env python
from setuptools import setup
setup(
  name = 'cs.deco',
  author = 'Cameron Simpson',
  author_email = 'cs@cskk.id.au',
  version = '20220311',
  url = 'https://bitbucket.org/cameron_simpson/css/commits/all',
  description =
    'Assorted decorator functions.',
  long_description =
    ('Assorted decorator functions.\n'    
 '\n'    
 '*Latest release 20220311*:\n'    
 '@cachedmethod: change the meaning of poll_delay=None to mean "never goes '    
 'stale" as I had thought it already did.\n'    
 '\n'    
 '## Function `cached(*args, **kwargs)`\n'    
 '\n'    
 'Former name for @cachedmethod.\n'    
 '\n'    
 '## Function `cachedmethod(*da, **dkw)`\n'    
 '\n'    
 'Decorator to cache the result of an instance or class method\n'    
 'and keep a revision counter for changes.\n'    
 '\n'    
 'The cached values are stored on the instance (`self`).\n'    
 'The revision counter supports the `@revised` decorator.\n'    
 '\n'    
 'This decorator may be used in 2 modes.\n'    
 'Directly:\n'    
 '\n'    
 '    @cachedmethod\n'    
 '    def method(self, ...)\n'    
 '\n'    
 'or indirectly:\n'    
 '\n'    
 '    @cachedmethod(poll_delay=0.25)\n'    
 '    def method(self, ...)\n'    
 '\n'    
 'Optional keyword arguments:\n'    
 '* `attr_name`: the basis name for the supporting attributes.\n'    
 '  Default: the name of the method.\n'    
 '* `poll_delay`: minimum time between polls; after the first\n'    
 '  access, subsequent accesses before the `poll_delay` has elapsed\n'    
 '  will return the cached value.\n'    
 '  Default: `None`, meaning the value never becomes stale.\n'    
 '* `sig_func`: a signature function, which should be significantly\n'    
 '  cheaper than the method. If the signature is unchanged, the\n'    
 '  cached value will be returned. The signature function\n'    
 '  expects the instance (`self`) as its first parameter.\n'    
 '  Default: `None`, meaning no signature function;\n'    
 '  the first computed value will be kept and never updated.\n'    
 '* `unset_value`: the value to return before the method has been\n'    
 '  called successfully.\n'    
 '  Default: `None`.\n'    
 '\n'    
 'If the method raises an exception, this will be logged and\n'    
 'the method will return the previously cached value,\n'    
 'unless there is not yet a cached value\n'    
 'in which case the exception will be reraised.\n'    
 '\n'    
 'If the signature function raises an exception\n'    
 'then a log message is issued and the signature is considered unchanged.\n'    
 '\n'    
 'An example use of this decorator might be to keep a "live"\n'    
 'configuration data structure, parsed from a configuration\n'    
 'file which might be modified after the program starts. One\n'    
 'might provide a signature function which called `os.stat()` on\n'    
 'the file to check for changes before invoking a full read and\n'    
 'parse of the file.\n'    
 '\n'    
 '*Note*: use of this decorator requires the `cs.pfx` module.\n'    
 '\n'    
 '## Function `contextdecorator(*da, **dkw)`\n'    
 '\n'    
 'A decorator for a context manager function `cmgrfunc`\n'    
 'which turns it into a decorator for other functions.\n'    
 '\n'    
 'This supports easy implementation of "setup" and "teardown"\n'    
 'code around other functions without the tedium of defining\n'    
 'the wrapper function itself. See the examples below.\n'    
 '\n'    
 'The resulting context manager accepts an optional keyword\n'    
 'parameter `provide_context`, default `False`. If true, the\n'    
 'context returned from the context manager is provided as the\n'    
 'first argument to the call to the wrapped function.\n'    
 '\n'    
 'Note that the context manager function `cmgrfunc`\n'    
 'has _not_ yet been wrapped with `@contextmanager`,\n'    
 'that is done by `@contextdecorator`.\n'    
 '\n'    
 'This decorator supports both normal functions and generator functions.\n'    
 '\n'    
 'With a normal function the process is:\n'    
 '* call the context manager with `(func,a,kw,*da,**dkw)`,\n'    
 '  returning `ctxt`,\n'    
 '  where `da` and `dkw` are the positional and keyword parameters\n'    
 '  supplied when the decorator was defined.\n'    
 '* within the context\n'    
 '  return the value of `func(ctxt,*a,**kw)` if `provide_context` is true\n'    
 '  or the value of `func(*a,**kw)` if not (the default)\n'    
 '\n'    
 'With a generator function the process is:\n'    
 '* obtain an iterator by calling `func(*a,**kw)`\n'    
 '* for iterate over the iterator, yielding its results,\n'    
 '  by calling the context manager with `(func,a,kw,**da,**dkw)`,\n'    
 '  around each `next()`\n'    
 'Note that it is an error to provide a true value for `provide_context`\n'    
 'if the decorated function is a generator function.\n'    
 '\n'    
 'Some examples follow.\n'    
 '\n'    
 'Trace the call and return of a specific function:\n'    
 '\n'    
 '    @contextdecorator\n'    
 '    def tracecall(func, a, kw):\n'    
 '        """ Trace the call and return from some function.\n'    
 '            This can easily be adapted to purposes such as timing a\n'    
 '            function call or logging use.\n'    
 '        """\n'    
 '        print("call %s(*%r,**%r)" % (func, a, kw))\n'    
 '        try:\n'    
 '          yield\n'    
 '        except Exception as e:\n'    
 '          print("exception from %s(*%r,**%r): %s" % (func, a, kw, e))\n'    
 '          raise\n'    
 '        else:\n'    
 '          print("return from %s(*%r,**%r)" % (func, a, kw))\n'    
 '\n'    
 '    @tracecall\n'    
 '    def f():\n'    
 '        """ Some function to trace.\n'    
 '        """\n'    
 '\n'    
 '    @tracecall(provide_context=True):\n'    
 '    def f(ctxt, *a, **kw):\n'    
 '        """ A function expecting the context object as its first argument,\n'    
 '            ahead of whatever other arguments it would normally require.\n'    
 '        """\n'    
 '\n'    
 "See who is making use of a generator's values,\n"    
 'when a generator might be invoked in one place and consumed elsewhere:\n'    
 '\n'    
 '    from cs.py.stack import caller\n'    
 '\n'    
 '    @contextdecorator\n'    
 '    def genuser(genfunc, *a, **kw):\n'    
 '        user = caller(-4)\n'    
 '        print(f"iterate over {genfunc}(*{a!r},**{kw!r}) from {user}")\n'    
 '        yield\n'    
 '\n'    
 '    @genuser\n'    
 '    def linesof(filename):\n'    
 '        with open(filename) as f:\n'    
 '            yield from f\n'    
 '\n'    
 '    # obtain a generator of lines here\n'    
 '    lines = linesof(__file__)\n'    
 '\n'    
 '    # perhaps much later, or in another function\n'    
 '    for lineno, line in enumerate(lines, 1):\n'    
 '        print("line %d: %d words" % (lineno, len(line.split())))\n'    
 '\n'    
 'Turn on "verbose mode" around a particular function:\n'    
 '\n'    
 '    import sys\n'    
 '    import threading\n'    
 '    from cs.context import stackattrs\n'    
 '\n'    
 '    class State(threading.local):\n'    
 '        def __init__(self):\n'    
 '            # verbose if stderr is on a terminal\n'    
 '            self.verbose = sys.stderr.isatty()\n'    
 '\n'    
 '    # per thread global state\n'    
 '    state = State()\n'    
 '\n'    
 '    @contextdecorator\n'    
 '    def verbose(func):\n'    
 '        with stackattrs(state, verbose=True) as old_attrs:\n'    
 "            if not old_attrs['verbose']:\n"    
 '                print(f"enabled verbose={state.verbose} for function '    
 '{func}")\n'    
 '            # yield the previous verbosity as the context\n'    
 "            yield old_attrs['verbose']\n"    
 '\n'    
 '    # turn on verbose mode\n'    
 '    @verbose\n'    
 '    def func(x, y):\n'    
 '        if state.verbose:\n'    
 '            # print if verbose\n'    
 '            print("x =", x, "y =", y)\n'    
 '\n'    
 '    # turn on verbose mode and also pass in the previous state\n'    
 '    # as the first argument\n'    
 '    @verbose(provide_context=True):\n'    
 '    def func2(old_verbose, x, y):\n'    
 '        if state.verbose:\n'    
 '            # print if verbose\n'    
 '            print("old_verbosity =", old_verbose, "x =", x, "y =", y)\n'    
 '\n'    
 '## Function `contextual(func)`\n'    
 '\n'    
 'Wrap a simple function as a context manager.\n'    
 '\n'    
 'This was written to support users of `@strable`,\n'    
 'which requires its `open_func` to return a context manager;\n'    
 'this turns an arbitrary function into a context manager.\n'    
 '\n'    
 'Example promoting a trivial function:\n'    
 '\n'    
 '    >>> f = lambda: 3\n'    
 '    >>> cf = contextual(f)\n'    
 '    >>> with cf() as x: print(x)\n'    
 '    3\n'    
 '\n'    
 '## Function `decorator(deco)`\n'    
 '\n'    
 'Wrapper for decorator functions to support optional arguments.\n'    
 '\n'    
 'The actual decorator function ends up being called as:\n'    
 '\n'    
 '    mydeco(func, *da, **dkw)\n'    
 '\n'    
 'allowing `da` and `dkw` to affect the behaviour of the decorator `mydeco`.\n'    
 '\n'    
 'Examples:\n'    
 '\n'    
 '    # define your decorator as if always called with func and args\n'    
 '    @decorator\n'    
 '    def mydeco(func, *da, arg2=None):\n'    
 '      ... decorate func subject to the values of da and arg2\n'    
 '\n'    
 '    # mydeco called with defaults\n'    
 '    @mydeco\n'    
 '    def func1(...):\n'    
 '      ...\n'    
 '\n'    
 '    @ mydeco called with nondefault arguments\n'    
 "    @mydeco('foo', arg2='bah')\n"    
 '    def func2(...):\n'    
 '      ...\n'    
 '\n'    
 '## Function `fmtdoc(func)`\n'    
 '\n'    
 "Decorator to replace a function's docstring with that string\n"    
 "formatted against the function's module `__dict__`.\n"    
 '\n'    
 'This supports simple formatted docstrings:\n'    
 '\n'    
 "    ENVVAR_NAME = 'FUNC_DEFAULT'\n"    
 '\n'    
 '    @fmtdoc\n'    
 '    def func():\n'    
 '        """Do something with os.environ[{ENVVAR_NAME}]."""\n'    
 '        print(os.environ[ENVVAR_NAME])\n'    
 '\n'    
 'This gives `func` this docstring:\n'    
 '\n'    
 '    Do something with os.environ[FUNC_DEFAULT].\n'    
 '\n'    
 '*Warning*: this decorator is intended for wiring "constants"\n'    
 'into docstrings, not for dynamic values. Use for other types\n'    
 'of values should be considered with trepidation.\n'    
 '\n'    
 '## Function `logging_wrapper(*da, **dkw)`\n'    
 '\n'    
 'Decorator for logging call shims\n'    
 'which bumps the `stacklevel` keyword argument so that the logging system\n'    
 'chooses the correct frame to cite in messages.\n'    
 '\n'    
 'Note: has no effect on Python < 3.8 because `stacklevel` only\n'    
 'appeared in that version.\n'    
 '\n'    
 '## Function `observable_class(property_names, only_unequal=False)`\n'    
 '\n'    
 'Class decorator to make various instance attributes observable.\n'    
 '\n'    
 'Parameters:\n'    
 '* `property_names`:\n'    
 '  an interable of instance property names to set up as\n'    
 '  observable properties. As a special case a single `str` can\n'    
 '  be supplied if only one attribute is to be observed.\n'    
 '* `only_unequal`:\n'    
 '  only call the observers if the new property value is not\n'    
 '  equal to the previous proerty value. This requires property\n'    
 '  values to be comparable for inequality.\n'    
 '  Default: `False`, meaning that all updates will be reported.\n'    
 '\n'    
 '## Function `OBSOLETE(*da, **dkw)`\n'    
 '\n'    
 'Decorator for obsolete functions.\n'    
 '\n'    
 'Use:\n'    
 '\n'    
 '    @OBSOLETE\n'    
 '    def func(...):\n'    
 '\n'    
 'This emits a warning log message before calling the decorated function.\n'    
 '\n'    
 '## Function `strable(*da, **dkw)`\n'    
 '\n'    
 'Decorator for functions which may accept a `str`\n'    
 'instead of their core type.\n'    
 '\n'    
 'Parameters:\n'    
 '* `func`: the function to decorate\n'    
 '* `open_func`: the "open" factory to produce the core type\n'    
 '  if a string is provided;\n'    
 '  the default is the builtin "open" function.\n'    
 '  The returned value should be a context manager.\n'    
 '  Simpler functions can be decorated with `@contextual`\n'    
 '  to turn them into context managers if need be.\n'    
 '\n'    
 'The usual (and default) example is a function to process an\n'    
 'open file, designed to be handed a file object but which may\n'    
 'be called with a filename. If the first argument is a `str`\n'    
 'then that file is opened and the function called with the\n'    
 'open file.\n'    
 '\n'    
 'Examples:\n'    
 '\n'    
 '    @strable\n'    
 '    def count_lines(f):\n'    
 '      return len(line for line in f)\n'    
 '\n'    
 '    class Recording:\n'    
 '      "Class representing a video recording."\n'    
 '      ...\n'    
 '    @strable(open_func=Recording)\n'    
 '    def process_video(r):\n'    
 '      ... do stuff with `r` as a Recording instance ...\n'    
 '\n'    
 '*Note*: use of this decorator requires the `cs.pfx` module.\n'    
 '\n'    
 '# Release Log\n'    
 '\n'    
 '\n'    
 '\n'    
 '*Release 20220311*:\n'    
 '@cachedmethod: change the meaning of poll_delay=None to mean "never goes '    
 'stale" as I had thought it already did.\n'    
 '\n'    
 '*Release 20220227*:\n'    
 '@cachedmethod: more paranoid access to the revision attribute.\n'    
 '\n'    
 '*Release 20210823*:\n'    
 '@decorator: preserve the __name__ of the wrapped function.\n'    
 '\n'    
 '*Release 20210123*:\n'    
 'Syntax backport for older Pythons.\n'    
 '\n'    
 '*Release 20201202*:\n'    
 '@decorator: tweak test for callable(da[0]) to accord with the docstring.\n'    
 '\n'    
 '*Release 20201025*:\n'    
 'New @contextdecorator decorator for context managers to turn them into '    
 'setup/teardown decorators.\n'    
 '\n'    
 '*Release 20201020*:\n'    
 '* @cachedmethod: bugfix cache logic.\n'    
 '* @strable: support generator functions.\n'    
 '\n'    
 '*Release 20200725*:\n'    
 'Overdue upgrade of @decorator to support combining the function and '    
 'decorator args in one call.\n'    
 '\n'    
 '*Release 20200517.2*:\n'    
 'Minor upgrade to @OBSOLETE.\n'    
 '\n'    
 '*Release 20200517.1*:\n'    
 'Tweak @OBSOLETE and @cached (obsolete name for @cachedmethod).\n'    
 '\n'    
 '*Release 20200517*:\n'    
 'Get warning() from cs.gimmicks.\n'    
 '\n'    
 '*Release 20200417*:\n'    
 '* @decorator: do not override __doc__ on the decorated function, just '    
 'provide default.\n'    
 '* New @logging_wrapper which bumps the `stacklevel` parameter in Python 3.8 '    
 'and above so that shims recite the correct caller.\n'    
 '\n'    
 '*Release 20200318.1*:\n'    
 'New @OBSOLETE to issue a warning on a call to an obsolete function, like an '    
 'improved @cs.logutils.OBSOLETE (which needs to retire).\n'    
 '\n'    
 '*Release 20200318*:\n'    
 '@cachedmethod: tighten up the "is the value changed" try/except.\n'    
 '\n'    
 '*Release 20191012*:\n'    
 '* New @contextual decorator to turn a simple function into a context '    
 'manager.\n'    
 '* @strable: mention context manager requirement and @contextual as '    
 'workaround.\n'    
 '\n'    
 '*Release 20191006*:\n'    
 'Rename @cached to @cachedmethod, leave compatible @cached behind which '    
 'issues a warning (will be removed in a future release).\n'    
 '\n'    
 '*Release 20191004*:\n'    
 'Avoid circular import with cs.pfx by removing requirement and doing the '    
 'import later if needed.\n'    
 '\n'    
 '*Release 20190905*:\n'    
 'Bugfix @deco: it turns out that you may not set the .__module__ attribute on '    
 'a property object.\n'    
 '\n'    
 '*Release 20190830.2*:\n'    
 'Make some getattr calls robust.\n'    
 '\n'    
 '*Release 20190830.1*:\n'    
 '@decorator: set the __module__ of the wrapper.\n'    
 '\n'    
 '*Release 20190830*:\n'    
 '@decorator: set the __module__ of the wrapper from the decorated target, '    
 'aids cs.distinf.\n'    
 '\n'    
 '*Release 20190729*:\n'    
 '@cached: sidestep uninitialised value.\n'    
 '\n'    
 '*Release 20190601.1*:\n'    
 '@strable: fix the example in the docstring.\n'    
 '\n'    
 '*Release 20190601*:\n'    
 '* Bugfix @decorator to correctly propagate the docstring of the '    
 'subdecorator.\n'    
 '* Improve other docstrings.\n'    
 '\n'    
 '*Release 20190526*:\n'    
 '@decorator: add support for positional arguments and rewrite - simpler and '    
 'clearer.\n'    
 '\n'    
 '*Release 20190512*:\n'    
 '@fmtdoc: add caveat against misuse of this decorator.\n'    
 '\n'    
 '*Release 20190404*:\n'    
 "New @fmtdoc decorator to format a function's doctsring against its module's "    
 'globals.\n'    
 '\n'    
 '*Release 20190403*:\n'    
 '* @cached: bugfix: avoid using unset sig_func value on first pass.\n'    
 '* @observable_class: further tweaks.\n'    
 '\n'    
 '*Release 20190322.1*:\n'    
 '@observable_class: bugfix __init__ wrapper function.\n'    
 '\n'    
 '*Release 20190322*:\n'    
 '* New class decorator @observable_class.\n'    
 '* Bugfix import of "warning".\n'    
 '\n'    
 '*Release 20190309*:\n'    
 '@cached: improve the exception handling.\n'    
 '\n'    
 '*Release 20190307.2*:\n'    
 'Fix docstring typo.\n'    
 '\n'    
 '*Release 20190307.1*:\n'    
 'Bugfix @decorator: final plumbing step for decorated decorator.\n'    
 '\n'    
 '*Release 20190307*:\n'    
 '* @decorator: drop unused arguments, they get used by the returned '    
 'decorator.\n'    
 '* Rework the @cached logic.\n'    
 '\n'    
 '*Release 20190220*:\n'    
 '* Bugfix @decorator decorator, do not decorate twice.\n'    
 "* Have a cut at inheriting the decorated function's docstring.\n"    
 '\n'    
 '*Release 20181227*:\n'    
 '* New decoartor @strable for function which may accept a str instead of '    
 'their primary type.\n'    
 '* Improvements to @cached.\n'    
 '\n'    
 '*Release 20171231*:\n'    
 'Initial PyPI release.'),
  install_requires = ['cs.gimmicks'],
  classifiers = ['Programming Language :: Python', 'Programming Language :: Python :: 2', 'Programming Language :: Python :: 3', 'Development Status :: 4 - Beta', 'Intended Audience :: Developers', 'Operating System :: OS Independent', 'Topic :: Software Development :: Libraries :: Python Modules', 'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)'],
  keywords = ['python2', 'python3'],
  license = 'GNU General Public License v3 or later (GPLv3+)',
  long_description_content_type = 'text/markdown',
  package_dir = {'': 'lib/python'},
  py_modules = ['cs.deco'],
)
